use std::{collections::BTreeMap, sync::Arc};

use tokio::sync::RwLock;

use crate::state::SharedState;

use super::{handler::Handler, staging::StagingEvent};

pub type ConsumerId = String;
pub type ConsumerIndex = String;
pub type Topic = String;
pub type PartitionKey = String;
pub type TopicOffset = usize;

// Consumer will at some point contain an interface to send events over a channel to a lib
#[derive(Debug, Clone)]
pub struct Consumer {
    pub offset: TopicOffset,

    pub tx: tokio::sync::mpsc::Sender<StagingEvent>,
    pub rx: Arc<tokio::sync::Mutex<tokio::sync::mpsc::Receiver<StagingEvent>>>,
}

impl Default for Consumer {
    fn default() -> Self {
        Self::new(0)
    }
}

impl Consumer {
    pub fn new(offset: usize) -> Self {
        let (tx, rx) = tokio::sync::mpsc::channel(128);

        Self {
            offset,
            tx,
            rx: Arc::new(tokio::sync::Mutex::new(rx)),
        }
    }
}

impl PartialEq for Consumer {
    fn eq(&self, other: &Self) -> bool {
        self.offset == other.offset
    }
}

#[derive(Clone)]
pub struct Consumers {
    storage: Arc<RwLock<BTreeMap<ConsumerId, BTreeMap<ConsumerIndex, Consumer>>>>,
    subscriptions: Arc<RwLock<BTreeMap<Topic, Vec<ConsumerId>>>>,
    handler: Handler,
}

// message arrives in queue
// message is stored in staging
// after staging, topic and key as a partition is notifing active consumers.
// active consumers now listen and receive events, if they process ok, the offset of a consumer will move ahead in the partition
// so we keep a system of record of each consumer
impl Consumers {
    pub fn new(handler: Handler) -> Self {
        Self {
            storage: Arc::default(),
            subscriptions: Arc::default(),
            handler,
        }
    }

    pub async fn add_consumer(
        &self,
        id: impl Into<ConsumerId>,
        index: impl Into<ConsumerIndex>,
        topic: impl Into<Topic>,
    ) -> anyhow::Result<()> {
        let id = id.into();
        let index = index.into();
        let topic = topic.into();

        {
            let mut storage = self.storage.write().await;

            if !storage.contains_key(&id) {
                storage.insert(id.clone(), BTreeMap::default());
            }

            let consumer_group = storage.get_mut(&id).unwrap();
            if !consumer_group.contains_key(&index) {
                consumer_group.insert(index.clone(), Consumer::default());
            }
        }

        {
            let mut subscriptions = self.subscriptions.write().await;
            if !subscriptions.contains_key(&topic) {
                subscriptions.insert(topic.clone(), Vec::default());
            }

            let subscription_consumers = subscriptions.get_mut(&topic).unwrap();
            if !subscription_consumers.contains(&id) {
                subscription_consumers.push(id.clone());
            }
        }

        Ok(())
    }

    pub async fn get_consumer(
        &self,
        id: impl Into<ConsumerId>,
        index: impl Into<ConsumerIndex>,
    ) -> Option<Consumer> {
        let storage = self.storage.read().await;
        let consumer_group = storage.get(&id.into())?;
        let consumer = consumer_group.get(&index.into())?;

        Some(consumer.to_owned())
    }

    pub async fn notify_update(
        &self,
        topic: impl Into<Topic>,
        key: Option<impl Into<PartitionKey>>,
        offset: impl Into<TopicOffset>,
    ) -> anyhow::Result<()> {
        let topic = topic.into();
        let offset = offset.into();
        let key = key.map(|k| k.into());

        let subscriptions = self.subscriptions.read().await;
        let subscription = match subscriptions.get(&topic) {
            Some(s) => s,
            None => {
                tracing::debug!(topic = &topic, "no subscription for topic");
                return Ok(());
            }
        };

        let mut storage = self.storage.write().await;
        for consumer_id in subscription {
            match storage.get_mut(consumer_id) {
                Some(consumer_groups) => {
                    for consumer in consumer_groups.values_mut() {
                        // TODO: Implement retry logic, etc.
                        self.handler
                            .handle_offset(&topic, key.as_ref(), consumer, offset)
                            .await?;
                        consumer.offset = offset;
                    }
                }
                None => {
                    tracing::trace!(
                        topic = &topic,
                        consumer_id = &consumer_id,
                        "found no consumer"
                    )
                }
            }
        }

        Ok(())
    }
}

pub trait ConsumersState {
    fn consumers(&self) -> Consumers;
}

impl ConsumersState for SharedState {
    fn consumers(&self) -> Consumers {
        self.consumers.clone()
    }
}

#[cfg(test)]
mod test {
    use tracing_test::traced_test;

    use crate::services::staging::{Staging, StagingEvent};

    use super::*;

    #[tokio::test]
    async fn can_add_consumer() -> anyhow::Result<()> {
        let consumer_id = "some-consumer-id";
        let consumer_index = "some-consumer-index";
        let topic = "some-topic";

        let consumers = Consumers::new(Handler::new(Staging::default()));

        consumers
            .add_consumer(consumer_id, consumer_index, topic)
            .await?;
        let consumer = consumers.get_consumer(consumer_id, consumer_index).await;

        assert_eq!(Some(Consumer::default()), consumer);

        Ok(())
    }

    #[tokio::test]
    #[traced_test]
    async fn can_notify_consumer() -> anyhow::Result<()> {
        let consumer_id = "some-consumer-id".to_string();
        let consumer_index = "some-consumer-index".to_string();
        let topic = "some-topic".to_string();
        let offset = 9usize;

        let staging = Staging::default();
        // Publish 10 messages
        for _ in 0..10 {
            let offset = staging
                .publish(StagingEvent {
                    topic: topic.clone(),
                    key: "".into(),
                    id: None,
                    published: chrono::Utc::now(),
                    value: Vec::new(),
                })
                .await?;

            tracing::trace!("published offset: {}", offset);
        }

        let consumers = Consumers::new(Handler::new(staging));

        consumers
            .add_consumer(&consumer_id, &consumer_index, &topic)
            .await?;
        let consumer = consumers.get_consumer(&consumer_id, &consumer_index).await;
        assert_eq!(Some(Consumer::default()), consumer);

        consumers
            .notify_update(&topic, None::<String>, offset)
            .await?;
        let consumer = consumers.get_consumer(&consumer_id, &consumer_index).await;
        assert_eq!(Some(Consumer::new(9)), consumer);

        Ok(())
    }
}
